#include "cs.h"                                 /*      AUXFD.C         */

static void fdchprint(INSDS*);

/* IV - Feb 22 2003 */

void auxalloc(long nbytes, AUXCH *auxchp)
     /* allocate an auxds, or expand an old one */
     /*    call only from init (xxxset) modules */
{
    void    *h = curip->aux_heap;
    size_t  n, m;

    n = (size_t) (nbytes + 3L) >> 2;
    if (h == NULL) {
      /* create auxalloc heap for current instrument if not done yet */
      h = curip->aux_heap =
        csoundHeapCreate((void*) &cglob, MEMALLOC_DISPOSABLE);
      if (h == NULL) {
        auxchp->size = 0L; auxchp->auxp = NULL; auxchp->endp = NULL;
        return;
      }
      /* set error code for handler */
      csoundHeapSetErrorCode(h, CSOUND_AUXALLOC_ERROR);
    }
    m = (size_t) (auxchp->size + 3L) >> 2;
    auxchp->size = nbytes;                      /* store size */
    if (auxchp->auxp != NULL) {
      /* if already allocated, */
      if (n == m) goto clear_space;     /* same size: clear only */
      /* different size: reallocate */
      auxchp->auxp = (char*) csoundRealloc(h, (void*) auxchp->auxp,
                                              sizeof(int32_t) * n);
    }
    else {
      /* not allocated yet, use malloc */
      auxchp->auxp = (char*) csoundMalloc(h, sizeof(int32_t) * n);
    }
    if (auxchp->auxp == NULL) {         /* safety */
      auxchp->size = 0L; auxchp->auxp = NULL; auxchp->endp = NULL;
      return;
    }
    auxchp->endp = auxchp->auxp + nbytes;       /* store end pointer */

 clear_space:
    /* clear allocated space to zero */
    memset((int32_t*) auxchp->auxp, 0, sizeof(int32_t) * n);
}

void fdrecord(FDCH *fdchp)  /* put fdchp into chain of fd's for this instr */
                            /*      call only from init (xxxset) modules   */
{
    FDCH        *prvchp, *nxtchp;

    prvchp = &curip->fdch;                      /* from current insds,  */
    while ((nxtchp = prvchp->nxtchp) != NULL)   /* chain through fdlocs */
      prvchp = nxtchp;
    prvchp->nxtchp = fdchp;                     /* then add this fdloc  */
    fdchp->nxtchp = NULL;                       /* & terminate the chain */
    VMSG(fdchprint(curip);)
}

void fdclose(FDCH *fdchp)       /* close a file and remove from fd chain */
                                /*  call only from inits, after fdrecord */
{
    FDCH        *prvchp, *nxtchp;

    prvchp = &curip->fdch;                      /* from current insds,  */
    while ((nxtchp = prvchp->nxtchp) != NULL) {     /* chain through fdlocs */
      if (nxtchp == fdchp) {            /*   till find this one */
        close(fdchp->fd);               /* then close the file  */
        fdchp->fd = 0;                  /*   delete the fd &    */
        prvchp->nxtchp = fdchp->nxtchp; /* unlnk from fdchain */
        VMSG(fdchprint(curip);)
          return;
      }
      else prvchp = nxtchp;
    }
    fdchprint(curip);
    sprintf(errmsg,Str("fdclose: no record of fd %d"), fdchp->fd);
    die(errmsg);
}

void fdchclose(INSDS *ip)   /* close all files in instr fd chain     */
                            /*   called by insert on deact & expire  */
{                           /*   (also musmon on s-code, & fgens for gen01) */
    FDCH        *curchp = &ip->fdch;
    int fd;

    VMSG(fdchprint(ip);)
      while ((curchp = curchp->nxtchp) != NULL) { /* for all fd's in chain: */
        if ((fd = curchp->fd) <= 2) {
          fdchprint(ip);
          sprintf(errmsg,Str("fdclose: illegal fd %d in chain"),fd);
          die(errmsg);
        }
        close(fd);                      /*      close the file  */
        curchp->fd = 0;                 /*      & delete the fd */
      }
      ip->fdch.nxtchp = NULL;           /* finally, delete the chain */
      VMSG(fdchprint(ip);)
}

static void fdchprint(INSDS *ip)   /* print the fd chain for this insds blk */
{
    FDCH        *curchp = &ip->fdch;

    printf(Str("fdlist for instr %d (%lx):"), ip->insno, ip);
    while ((curchp = curchp->nxtchp) != NULL)    /* chain through fdlocs */
      printf(Str("  fd %d in %lx"), curchp->fd, curchp);
    printf("\n");
}

