                                                                /* SYSDEP.H */
#ifndef CSOUND_SYSDEP_H
#define CSOUND_SYSDEP_H

#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#define PUBLIC          __declspec(dllexport)
#define LIBRARY_CALL    WINAPI
#else
#define PUBLIC
#define LIBRARY_CALL
#endif

#ifdef __cplusplus
extern "C" {
#endif
                                /* Experiment with doubles or floats */
#ifndef __CSND_FL_DEF
#  ifdef MYFLT_IS_FLOAT
#    undef      MYFLT_IS_FLOAT
#  endif
#  ifdef MYFLT_IS_DOUBLE
#    undef      MYFLT_IS_DOUBLE
#  endif
#  ifndef MYFLT
#    define     MYFLT   float
#    define     FL(x)   x##f
#    define     MYFLT_IS_FLOAT  1
#  else
#    if defined(float) || defined(double)
#      error "float or double is already defined as a macro"
#    else
#      define   float   4
#      define   double  8
#      if (MYFLT == float)
#        define FL(x)   x##f
#        define MYFLT_IS_FLOAT  1
#      elif (MYFLT == double)
#        define FL(x)   x
#        define MYFLT_IS_DOUBLE 1
#      else
#        error "Unknown MYFLT type. Must be defined to float or double."
#      endif
#      undef    float
#      undef    double
#    endif
#  endif
#  define __CSND_FL_DEF 1
#endif

#if defined(macintosh)
#  define   mac_classic     /* All Mac Compiles Before OSX, including Carbon */
#  define   mills_macintosh /* DEFINE THIS to COMPILE the Mills"Perf"Version */
#  include <stdlib.h>
#  include <unistd.h>       /* for open() etc protos on mac */
#  include <stat.h>
#  define   SYS5
#  define   SFDIGDES
#  define   WINDOWS
#  define   u_char  unsigned char
#  define   u_short unsigned short
#  define   u_int   unsigned int
#  define   u_long  unsigned long
#  define   O_NDELAY (0)
#  ifdef mills_macintosh
#    define exit(x) die("");
#  endif
#  define DIRSEP ':'
#elif defined(SYMANTEC)
#  include <stdlib.h>
#  include <unix.h>     /* for open() etc protos on mac */
#  define   SYS5
#  define   SFDIGDES
#  define   WINDOWS     /* with winmac.c */
#  define   u_char  unsigned char
#  define   u_short unsigned short
#  define   u_int   unsigned int
#  define   u_long  unsigned long
#  define   DIRSEP ':'
   extern off_t lseek(int, off_t, int);
#else
#  if defined(LINUX)
#    include <unistd.h>         /* IV - Jan 30 2003 */
#  endif
#  define DIRSEP '/'
#  ifdef LATTICE
#    define  u_char  unsigned char
#    define  u_short unsigned short
#    define  u_int   unsigned int
#    define  u_long  unsigned long
#    include <fcntl.h>
#    include <sys/types.h>
#  else
#    ifdef __WATCOMC__
#      define  u_char  unsigned char
#      define  u_short unsigned short
#      define  u_int   unsigned int
#      define  u_long  unsigned long
#      define  O_NDELAY (0)
#      include <io.h>
#     else
#       ifdef WIN32
#         define  u_char  unsigned char
#         define  u_short unsigned short
#         define  u_int   unsigned int
#         define  u_long  unsigned long
#         define  O_NDELAY (0)          /* IV - Oct 26 2002 */
#         ifdef __CYGWIN__
#           include <unistd.h>
#         endif
#         include <io.h>
#         include <fcntl.h>
#         undef   DIRSEP
#         define  DIRSEP    '\\'
#       else
#         ifdef DOSGCC
#           define O_NDELAY (0)
#           undef   DIRSEP
#           define  DIRSEP    '\\'
#         endif
#         include <sys/types.h>
#       endif
/*  RWD for WIN32 on VC++ */
#       ifndef _MSC_VER
#         include <sys/file.h>
#       endif
#     endif
#  endif
#  include <sys/stat.h>
#endif

#ifdef SYS5
# ifndef index
# define  index(A,B) strchr(A,B)
# endif
# include <fcntl.h>
# include <string.h>
#else
# ifdef WIN32
#  include <string.h>
#  include <fcntl.h>
# else
#  include <strings.h>
# endif
#endif

#ifdef __BEOS__
# include <unistd.h>
# if !defined(O_NDELAY) && !defined(CSSVINTERFACE_H)
#  error "Please use Makefile.be to build the BeOS version of Csound."
# endif
#endif

/* IV - Feb 08 2003: fixed size integer types */

#if defined(__GNUC__) && (__GNUC__ >= 3)
#  include <stdint.h>
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)
/* use standard header if available */
#  include <stdint.h>
#elif !defined(__int8_t_defined)
    /* these should be fairly portable, */
    typedef     signed char     int8_t;
    typedef     unsigned char   uint8_t;
    typedef     short           int16_t;
    typedef     unsigned short  uint16_t;
    typedef     int             int32_t;
    typedef     unsigned int    uint32_t;
#  ifdef __GNUC__
        /* unlike the following ones */
        typedef     long long           int64_t;
        typedef     unsigned long long  uint64_t;
#  elif defined(MSVC) || defined(_MSC_VER)
        typedef     __int64             int64_t;
        typedef     unsigned __int64    uint64_t;
#  endif
    /* these do not work on Win64, but are OK on 64 bit UNIX */
    /* systems and all 32 bit platforms */
    typedef     long            intptr_t;
    typedef     unsigned long   uintptr_t;
#endif

#ifndef PI
#define PI      (3.14159265358979323846)
#endif
#define TWOPI   (6.28318530717958647693)
#define PI_F    (FL(3.14159265358979323846))
#define TWOPI_F (FL(6.28318530717958647693))

/* IV - Feb 22 2003: new macros for handling denormals */
/* (N.B. with GCC these seem to require -fno-strict-aliasing) */

#if defined(__i386__) && defined(NO_DENORMALS)

/* Set float variable "x" to zero if it is a denormal. */
/* Note: x must be a variable, and not a constant or an expression. */

#define UNDENORMALIZE_FLOAT(x)                                          \
{                                                                       \
    register uint32_t *p = (uint32_t*) &(x);                            \
    if (!(*p & (uint32_t) 0x7F800000)) *p = (uint32_t) 0;               \
}

/* Set double variable "x" to zero if it is a denormal. */

#define UNDENORMALIZE_DOUBLE(x)                                         \
{                                                                       \
    register uint32_t *p = (uint32_t*) &(x);                            \
    if (!(p[1] & (uint32_t) 0x7FF00000)) p[0] = p[1] = (uint32_t) 0;    \
}

/* Set MYFLT variable "x" to zero if it is a denormal. */

#ifdef MYFLT_IS_FLOAT
#define UNDENORMALIZE(x)    UNDENORMALIZE_FLOAT(x)
#else
#define UNDENORMALIZE(x)    UNDENORMALIZE_DOUBLE(x)
#endif

#else           /* __i386__ && NO_DENORMALS */

/* Dummy macros for non-x86 systems or when NO_DENORMALS is not defined. */

#define UNDENORMALIZE_FLOAT(x)
#define UNDENORMALIZE_DOUBLE(x)
#define UNDENORMALIZE(x)

#endif          /* __i386__ && NO_DENORMALS */

#ifdef __cplusplus
}
#endif

#endif          /* CSOUND_SYSDEP_H */

