                                                        /*      GLOBALS.H   */
#ifndef CSOUND_GLOBALS_H
#define CSOUND_GLOBALS_H

/* globals.h -- written by Istvan Varga, 2003 */

/* for MYFLT */
#include "Csound/sysdep.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef CSOUND_API_HEADERS
#error "CsoundAPI applications must not access GLOBALS structure directly"
#endif

typedef struct GLOBALS_
{
    /* Csound API functions. This is the only public part of the GLOBALS   */
    /* structure, and should always be the first so that user applications */
    /* can cast to CSOUNDAPI and access the functions.                     */
    CSOUNDAPI       api_functions;

    /* private global variables: not to be accessed from plugins */
    /* and API applications */
    OPARMS          O;
    CSGLOBALS       csglobals;
    /* orchestra header variables calculated by otran */
    MYFLT           tran_sr, tran_kr;
    MYFLT           tran_ksmps;
    MYFLT           tran_0dbfs;
    int             tran_nchnls;
    /* orchestra / score names */
    char            * orchname, * scorename;
    char            * xfilename;
    /* opcode list, user defined opcodes, and named instruments */
    OENTRY          * opcodlst;
    void            * opcode_list;              /* IV - Oct 31 2002 */
    OENTRY          * oplstend;
    OPCODINFO       * opcodeInfo;               /* IV - Oct 20 2002 */
    void            * instrumentNames;
    /* 0dBFS system */
    MYFLT           dbfs_to_short;
    MYFLT           short_to_dbfs;
    MYFLT           dbfs_to_float;
    MYFLT           float_to_dbfs;
    MYFLT           dbfs_to_long;
    MYFLT           long_to_dbfs;
    /* real-time audio device numbers for Win32 */
    unsigned int    rtin_dev;
    unsigned int    rtout_dev;
    /* audio I/O buffers (inbufsamps/outbufsamps samples in any format) */
    void            *inbuf;
    void            *outbuf;
    /* for 'midiin' opcode */
    int             MIDIINbufIndex;
    MIDIMESSAGE     MIDIINbuffer2[MIDIINBUFMAX];
    /* peak amplitude and position, and out of range samples */
    MYFLT           maxamp[MAXCHNLS];           /* note */
    MYFLT           smaxamp[MAXCHNLS];          /* section */
    MYFLT           omaxamp[MAXCHNLS];          /* overall */
    MYFLT           * maxampend;
    unsigned long   maxpos[MAXCHNLS];
    unsigned long   smaxpos[MAXCHNLS];
    unsigned long   omaxpos[MAXCHNLS];
    long            rngcnt[MAXCHNLS];
    short           rngflg;
    short           multichan;                  /* non-zero if nchnls > 1 */
    /* oload.h (???) */
    short           nlabels;
    short           ngotos;
    int             peakchunks;
    /* misc. global variables (FIXME: to be documented ... ) */
    int             csMagic1;
    EVTBLK          * currevent;
    int             keep_tmp;
    int             dither_output;
    FILE            * dribble;                  /* log file */
    int32_t         holdrand;
    int             maxinsno;
    int             maxopcno;                   /* IV - Oct 24 2002 */
    EVTBLK          * Linevtblk;
    long            nrecs;
    FILE            * Linepipe;
    int             Linefd;
    MYFLT           * ls_table;
    MYFLT           curr_func_sr;
    char            * retfilnam;
    INSTRTXT        ** instrtxtp;
    /* sprintf space for compiling msgs */
    char            errmsg[ERRSIZ];
    FILE            * scfp;
    FILE            * oscfp;
    char            * ssdirpath, * sfdirpath;
    char            * tokenstring;
    POLISH          * polish;
    FILE            * scorein;
    FILE            * scoreout;
    MYFLT           * pool;
    int             * argoffspace;              /* IV - Feb 22 2003 */
    INSDS           * frstoff;
    int             sensType;
    SRTBLK          * frstbp;
    int             sectcnt;
    long            section_start_kcnt;         /* IV - Feb 18 2003 */
    MCHNBLK         * m_chnbp[MAXCHAN];
    MYFLT           * cpsocint, * cpsocfrc;
    int             inerrcnt;                   /* init error count */
    int             synterrcnt;                 /* syntax error count */
    int             perferrcnt;                 /* perf error count */
    int             MIDIoutDONE;
    int             midi_out;
    char            strmsg[100];
    INSTRTXT        instxtanchor;
    INSDS           actanchor;
    /* List of started events, used in playevents() */
    EVTNODE         OrcTrigEvts;
    char            name_full[256];             /* Remember name used */
    int             Mforcdecs, Mxtroffs;
    int             MTrkend;
    char            * sadirpath;
    char            * oplibs;
    OPARMS          * oparms;
    jmp_buf         exitjmp;
    /* pointer to user data */
    void            * hostdata;
    int             csMagic2;
    /* IV - Feb 22 2003 */
    AUDIOPARMS      rtin_params;
    AUDIOPARMS      rtout_params;
    void            *audioInData;
    void            *audioOutData;
} GLOBALS;

/* Reset all global variables of a Csound instance, and         */
/* install default CsoundAPI callback functions on first call.  */

void globalsRESET(void*);

/* Same as globalsRESET(), but has no effect  */
/* if globals are already initialised.        */

void globals_init(void*);

#ifdef __cplusplus
}
#endif

#endif          /* CSOUND_GLOBALS_H */

