
/* main.c -- written by Istvan Varga, 2003 */

#include <cs.h>
#include <ctype.h>

extern  void    remove_tmpfiles(void);
extern  void    set_rt_priority(int, char**);

static  int     run_cnt = 1;         /* times to run */

/* select language of messages based on environment variable "CS_LANG"      */
/* currently, the allowed choices are 'de', 'en', 'es', 'fr', 'it', 'sv',   */
/* 'uk', and 'us' (case is not significant)                                 */

static cslanguage_t select_language(void)
{
    char    *s, tmp[256];
    int     i;

    s = getenv("CS_LANG");
    if (s == NULL)
      return CSLANGUAGE_DEFAULT;
    for (i = 0; s[i] != '\0'; i++) {
      if (isupper(s[i]))
        tmp[i] = tolower(s[i]);
      else
        tmp[i] = s[i];
    }
    tmp[i] = '\0';
    if (strcmp(tmp, "de") == 0)
      return CSLANGUAGE_GERMAN;
    else if (strcmp(tmp, "en") == 0)
      return CSLANGUAGE_ENGLISH_UK;
    else if (strcmp(tmp, "es") == 0)
      return CSLANGUAGE_SPANISH;
    else if (strcmp(tmp, "fr") == 0)
      return CSLANGUAGE_FRENCH;
    else if (strcmp(tmp, "it") == 0)
      return CSLANGUAGE_ITALIAN;
    else if (strcmp(tmp, "sv") == 0)
      return CSLANGUAGE_SWEDISH;
    else if (strcmp(tmp, "uk") == 0)
      return CSLANGUAGE_ENGLISH_UK;
    else if (strcmp(tmp, "us") == 0)
      return CSLANGUAGE_ENGLISH_US;

    return CSLANGUAGE_DEFAULT;
}

int main(int argc, char **argv)
{
    void  *csnd;
    int   retval;

    /* Real-time priority on Linux by Istvan Varga (Jan 6 2002) */
    /* This function is called before anything else to avoid    */
    /* running "normal" Csound code with setuid root.           */
    /* set_rt_priority gives up root privileges after setting   */
    /* priority and locking memory; init_getstring () and other */
    /* functions below will be run as a normal user (unless     */
    /* Csound was actually run as root, and not setuid root).   */

#if defined(LINUX)
    set_rt_priority(argc, argv);
#endif
    atexit(remove_tmpfiles);
    /* set language of messages */
    csoundSetLanguage(select_language());
    /* debugging flag for testing re-entrant operation */
    if (argc > 1) {
      if (!strncmp(argv[1], "--reentry-cnt=", 14)) {
        sscanf(argv[1] + 14, "%d", &run_cnt);
        if (run_cnt < 1) run_cnt = 1;
        /* skip option */
        argc--; argv++;
      }
    }
    do {
      /* initialise */
      csnd = csoundCreate(NULL);
      if (csnd == NULL)
        return -1;
      /* perform */
      retval = csoundPerform(csnd, argc, argv);
      if (retval)       /* in case of error:     */
        return retval;  /* terminate immediately */
      /* delete instance */
      csoundDestroy(csnd);
    } while (--run_cnt);

    return 0;
}

