
/* mstubs.c -- written by Istvan Varga, Feb 2003 */

/* implementation of mmalloc, mcalloc, mrealloc, mfree, and memRESET */
/* for compatibility with existing Csound sources                    */

#include <cs.h>

/* setting this to 1 allows freeing all memory allocated by Csound */
/* on csoundReset(), but makes memory allocations slower           */

#define MEMRESET_ALL_FREE   0

#if MEMRESET_ALL_FREE

static void *heap_handle = NULL;

#define INIT_HEAP                                                       \
{                                                                       \
    if (heap_handle == NULL) {                                          \
      heap_handle =                                                     \
        csoundHeapCreate((void*) &cglob, MEMALLOC_DISPOSABLE);          \
    }                                                                   \
}

void memRESET(void)
{
    rlsmemfiles();
    csoundHeapDestroy(&heap_handle);
}

char *mcalloc(long nbytes)
{
    INIT_HEAP
    return (char*) csoundCalloc(heap_handle, (size_t) nbytes, (size_t) 1);
}

char *mmalloc(long nbytes)
{
    INIT_HEAP
    return (char*) csoundMalloc(heap_handle, (size_t) nbytes);
}

void *mrealloc(void *old, long nbytes)
{
    INIT_HEAP
    return (void*) csoundRealloc(heap_handle, (void*) old, (size_t) nbytes);
}

void mfree(void *ptr)
{
    INIT_HEAP
    csoundFree(heap_handle, (void*) ptr);
}

#else           /* MEMRESET_ALL_FREE */

void memRESET(void)
{
    rlsmemfiles();
}

char *mcalloc(long nbytes)
{
    void  *p;

    p = (void*) calloc((size_t) nbytes, (size_t) 1);
    if (p == NULL) {
      ((CSOUNDAPI*) &cglob)->Die((void*) &cglob, "csound: not enough memory");
      return NULL;
    }
    return ((char*) p);
}

char *mmalloc(long nbytes)
{
    void  *p;

    p = (void*) malloc((size_t) nbytes);
    if (p == NULL) {
      ((CSOUNDAPI*) &cglob)->Die((void*) &cglob, "csound: not enough memory");
      return NULL;
    }
    return ((char*) p);
}

void *mrealloc(void *old, long nbytes)
{
    void  *p;

    if (nbytes == 0L) {
      if (old != NULL) free(old);
      return NULL;
    }
    else if (old == NULL) {
      p = (void*) malloc((size_t) nbytes);
    }
    else {
      p = (void*) realloc((void*) old, (size_t) nbytes);
    }
    if (p == NULL) {
      ((CSOUNDAPI*) &cglob)->Die((void*) &cglob, "csound: not enough memory");
      return NULL;
    }
    return ((void*) p);
}

void mfree(void *ptr)
{
    if (ptr != NULL)
      free(ptr);
}

#endif          /* !MEMRESET_ALL_FREE */

