
#include <Csound/csdl.h>
#include <math.h>
#include "vectors.h"

#define randGab         ENGINE->RandomValueU(p)
#define BiRandGab       ENGINE->RandomValueB(p)

static void vectorOp_set(VECTOROP *p)
{
    FUNC    *ftp;
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifn)) != NULL) {
      p->vector = ftp->ftable;
      p->elements = (int) *p->ielements;
    }
    else
      return;
    if ( p->elements > ftp->flen )
      ENGINE->InitError((void*) ENGINE, "vectorop: invalid num of elements");
}

static void vadd(VECTOROP *p)
{
    int     elements = p->elements;
    MYFLT   *vector = p->vector;
    MYFLT   value = *p->kval;

    do {
      *vector++ += value;
    } while (--elements);
}

static void vmult(VECTOROP *p)
{
    int elements = p->elements;
    MYFLT *vector = p->vector;
    MYFLT value = *p->kval;

    do      *vector++ *= value;
    while (--elements);
}

static void vpow(VECTOROP *p)
{
    int     elements = p->elements;
    MYFLT   *vector = p->vector;
    MYFLT   value = *p->kval;

    do {
      *vector = (MYFLT) pow((double) *vector, (double) value);
      vector++;
    } while (--elements);
}

static void vexp(VECTOROP *p)
{
    int     elements = p->elements;
    MYFLT   *vector = p->vector;
    MYFLT   value = *p->kval;

    do {
      *vector = (MYFLT) pow((double) value, (double) *vector);
      vector++;
    } while (--elements);
}

/*-------------------------*/

static void vectorsOp_set(VECTORSOP *p)
{
    FUNC    *ftp1, *ftp2;
    if ((ftp1 = ENGINE->QueryFTableInit((void*) p, p->ifn1)) != NULL) {
      p->vector1 = ftp1->ftable;
    }
    else
      return;
    if ((ftp2 = ENGINE->QueryFTableInit((void*) p, p->ifn2)) != NULL) {
      p->vector2 = ftp2->ftable;
    }
    else
      return;
    p->elements = (int) *p->ielements;
    if ( p->elements > ftp1->flen || p->elements > ftp2->flen)
      ENGINE->InitError((void*) ENGINE, "vectorop: invalid num of elements");
}


static void vcopy(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do      *vector1++ = *vector2++;
    while (--elements);
}


static void vaddv(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do      *vector1++ += *vector2++;
    while (--elements);
}

static void vsubv(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do      *vector1++ -= *vector2++;
    while (--elements);
}

static void vmultv(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do      *vector1++ *= *vector2++;
    while (--elements);
}

static void vdivv(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do      *vector1++ /= *vector2++;
    while (--elements);
}

static void vpowv(VECTORSOP *p)
{
    int     elements = p->elements;
    MYFLT   *vector1 = p->vector1, *vector2 = p->vector2;

    do {
      *vector1 = (MYFLT) pow((double) *vector1, (double) *vector2);
      vector1++;
      vector2++;
    } while (--elements);
}

static void vexpv(VECTORSOP *p)
{
    int     elements = p->elements;
    MYFLT   *vector1 = p->vector1, *vector2 = p->vector2;

    do {
      *vector1 = (MYFLT) pow((double) *vector2, (double) *vector1);
      vector1++;
      vector2++;
    } while (--elements);
}

static void vmap(VECTORSOP *p)
{
    int elements = p->elements;
    MYFLT *vector1 = p->vector1, *vector2 = p->vector2;

    do {
      *vector1 = vector2[(int) (*vector1 + FL(0.01))];
      vector1++;
      vector2++;
    } while (--elements);
}

static void vlimit_set(VLIMIT *p)
{
    FUNC    *ftp;
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifn)) != NULL) {
      p->vector = ftp->ftable;
      p->elements = (int) *p->ielements;
    }
    else
      return;
    if (p->elements >= ftp->flen)
      ENGINE->InitError((void*) ENGINE, "vectorop: invalid num of elements");
}

static void vlimit(VLIMIT *p)
{
    int     elements = p->elements;
    MYFLT   *vector = p->vector;
    MYFLT   min = *p->kmin, max = *p->kmax;
    do {
      *vector = (*vector > min) ? ((*vector < max) ? *vector : max) : min;
      vector++;
    } while (--elements);
}

/*-----------------------------*/
static void vport_set(VPORT *p)
{
    FUNC    *ftp;
    int     elements;
    MYFLT   *vector, *yt1, *vecInit = NULL;

    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifn)) != NULL) {
      vector = (p->vector = ftp->ftable);
      elements = (p->elements = (int) *p->ielements);
      if (elements >= ftp->flen) {
        ENGINE->InitError((void*) ENGINE,
                          "vport: invalid table length or num of elements");
        return;
      }
    }
    else {
      ENGINE->InitError((void*) ENGINE, "vport: invalid table");
      return;
    }
    if (*p->ifnInit > FL(0.25)) {
      if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifnInit)) != NULL) {
        vecInit = ftp->ftable;
        if (elements >= ftp->flen) {
          ENGINE->InitError((void*) ENGINE,
            "vport: invalid init table length or num of elements");
          return;
        }
      }
      else {
        ENGINE->InitError((void*) ENGINE, "vport: invalid init table");
        return;
      }
    }
    if (p->auxch.auxp == NULL ||
        p->auxch.size < (long) (elements * sizeof(MYFLT)))
      ENGINE->AuxAlloc((void*) p, &p->auxch,
                       (size_t) (elements * sizeof(MYFLT)));
    yt1 = (p->yt1 = (MYFLT *) p->auxch.auxp);
    if (vecInit) {
      memcpy(yt1, vecInit, sizeof(MYFLT) * elements);
    } else {
      memset(yt1, 0, sizeof(MYFLT) * elements);
    }
    p->prvhtim = FL(-1.12123e35);
}

static void vport(VPORT *p)
{
    int elements = p->elements;
    MYFLT *vector = p->vector, *yt1 = p->yt1, c1, c2;

    if (p->prvhtim != *p->khtim) {
      p->c2 = (MYFLT) pow(0.5, (double) GLOBLS->onedkr / (double) *p->khtim);
      p->c1 = FL(1.0) - p->c2;
      p->prvhtim = *p->khtim;
    }
    c1 = p->c1;
    c2 = p->c2;
    do {
      *vector = c1 * *vector + c2 * *yt1;
      *yt1 = *vector;
      vector++;
      yt1++;
    } while (--elements);
}

/* ------------------------------- */

static void vwrap(VLIMIT *p)
{
    int elements = p->elements;
    MYFLT *vector = p->vector;
    MYFLT min = *p->kmin, max = *p->kmax;

    if (min >= max) {
      MYFLT average = (min + max) * FL(0.5);
      do {
        *vector++ = average;
      } while (--elements);
    }
    else {
      do {
        if (*vector >= max)
          *vector = min + (MYFLT) fmod(*vector - min, max - min);
        else if (*vector <= min)
          *vector = max - (MYFLT) fmod(max - *vector, max - min);
        vector++;
      } while (--elements);
    }
}

static void vmirror(VLIMIT *p)
{
    int elements = p->elements;
    MYFLT *vector = p->vector;
    MYFLT min = *p->kmin, max = *p->kmax;

    if (min >= max) {
      MYFLT average = (min+max)/2;
      do *vector++ = average;
      while (--elements);
    }
    else {
      do {
        while (!((*vector <= max) && (*vector >= min))) {
          if (*vector > max)
            *vector = max + max - *vector;
          else
            *vector = min + min - *vector;
        }
        vector++;
      } while (--elements);
    }
}

static void vrandh_set(VRANDH *p)
{
    FUNC    *ftp;
    int     elements;
    MYFLT   *num1;

    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifn)) != NULL) {
      p->vector = ftp->ftable;
      elements = (p->elements = (int) *p->ielements);
    }
    else
      return;
    if (p->elements >= ftp->flen) {
      ENGINE->InitError((void*) ENGINE, "vectorop: invalid num of elements");
      return;
    }
    p->phs = 0;
    if (p->auxch.auxp == NULL ||
        p->auxch.size < (long) (p->elements * sizeof(MYFLT)))
      ENGINE->AuxAlloc((void*) p, &p->auxch,
                       (size_t) (p->elements * sizeof(MYFLT)));
    num1 = (p->num1 = (MYFLT *) p->auxch.auxp);
    do {
      *num1++ = BiRandGab;
    } while (--elements);
}

static void vrandh(VRANDH *p)
{
    MYFLT   *vector = p->vector, *num1 = p->num1;
    MYFLT   value = *p->krange;
    int     elements = p->elements;

    do {
      *vector++ += *num1++ * value;
    } while (--elements);

    p->phs += (long)(*p->kcps * GLOBLS->kicvt);
    if (p->phs >= MAXLEN) {
      p->phs    &= PHMASK;
      elements  = p->elements;
      vector    = p->vector;
      num1      = p->num1;
      do {
        *num1++ = BiRandGab;
      } while (--elements);
    }
}

static void vrandi_set(VRANDI *p)
{
    FUNC    *ftp;
    int     elements;
    MYFLT   *dfdmax, *num1, *num2;

    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifn)) != NULL) {
      p->vector = ftp->ftable;
      elements = (p->elements = (int) *p->ielements);
    }
    else
      return;
    if (p->elements >= ftp->flen) {
      ENGINE->InitError((void*) ENGINE, "vectorop: invalid num of elements");
      return;
    }
    p->phs = 0;
    if (p->auxch.auxp == NULL)
      ENGINE->AuxAlloc((void*) p, &p->auxch,
                       (size_t) (elements * sizeof(MYFLT) * 3));
    num1 = (p->num1 = (MYFLT *) p->auxch.auxp);
    num2 = (p->num2 = &num1[elements]);
    dfdmax = (p->dfdmax = &num1[elements * 2]);

    do {
      *num1 = 0.f;
      *num2 = BiRandGab;
      *dfdmax++ = (*num2++ - *num1++) / FMAXLEN;
    } while (--elements);
}


static void vrandi(VRANDI *p)
{
    MYFLT   *vector = p->vector, *num1 = p->num1, *num2, *dfdmax = p->dfdmax;
    MYFLT   value = *p->krange;
    int     elements = p->elements;

    do *vector++ += (*num1++ + (MYFLT)p->phs * *dfdmax++) * value;
    while (--elements);

    p->phs += (long)(*p->kcps * GLOBLS->kicvt);
    if (p->phs >= MAXLEN) {
      p->phs    &= PHMASK;
      elements  = p->elements;
      vector    = p->vector;
      num1      = p->num1;
      num2      = p->num2;
      dfdmax    = p->dfdmax;

      do {
        *num1 = *num2;
        *num2 = BiRandGab ;
        *dfdmax++ = (*num2++ - *num1++) / FMAXLEN;
      } while (--elements);
    }
}

/* ----------------- */

static void vecdly_set(VECDEL *p)
{
    FUNC    *ftp;
    int     elements = (p->elements = (int) *p->ielements), j;
    long    n;

    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifnOut)) != NULL) {
      p->outvec = ftp->ftable;
      elements = (p->elements = (int) *p->ielements);
      if (elements >= ftp->flen) {
        ENGINE->InitError((void*) ENGINE, "vecdelay: invalid num of elements");
        return;
      }
    }
    else {
      ENGINE->InitError((void*) ENGINE, "vecdly: invalid output table");
      return;
    }
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifnIn)) != NULL) {
      p->invec = ftp->ftable;
      if (elements >= ftp->flen) {
        ENGINE->InitError((void*) ENGINE, "vecdelay: invalid num of elements");
        return;
      }
    }
    else {
      ENGINE->InitError((void*) ENGINE, "vecdly: invalid input table");
      return;
    }
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ifnDel)) != NULL) {
      p->dlyvec = ftp->ftable;
      if (elements >= ftp->flen) {
        ENGINE->InitError((void*) ENGINE, "vecdelay: invalid num of elements");
        return;
      }
    }
    else {
      ENGINE->InitError((void*) ENGINE, "vecdly: invalid delay table");
      return;
    }

    n = (p->maxd = (long) (*p->imaxd * GLOBLS->ekr));
    if (n == 0) n = (p->maxd = 1);

    if (!*p->istod) {
      size_t    bufptr_size, buf_size, bufleft_size;
      bufptr_size = ((size_t) elements * sizeof(MYFLT*) + (size_t) 7)
                    & (~((size_t) 7));
      buf_size = ((size_t) (n * elements) * sizeof(MYFLT) + (size_t) 7)
                    & (~((size_t) 7));
      bufleft_size = ((size_t) elements * sizeof(long) + (size_t) 7)
                    & (~((size_t) 7));
      if (p->aux.auxp == NULL ||
          (long) (bufptr_size + buf_size + bufleft_size) > p->aux.size) {
        ENGINE->AuxAlloc((void*) p, &p->aux,
                         bufptr_size + buf_size + bufleft_size);
        p->buf = (MYFLT **) p->aux.auxp;
        for (j = 0; j < elements; j++) {
          p->buf[j] = (MYFLT*) ((uint8_t*) p->aux.auxp
                                + (long) (bufptr_size + sizeof(MYFLT) * n * j));
        }
        p->left = (long*) ((uint8_t*) p->aux.auxp
                           + (long) (bufptr_size + buf_size));
      }
      else {
        for (j = 0; j < elements; j++) {
          memset(p->buf[j], 0, n * sizeof(MYFLT));
          p->left[j] = 0L;
        }
      }
    }
}


static void vecdly(VECDEL *p)
{
    long    maxd = p->maxd, *indx=p->left, v1, v2;
    MYFLT   **buf = p->buf, fv1, fv2, *inVec = p->invec;
    MYFLT   *outVec = p->outvec, *dlyVec = p->dlyvec;
    int     elements = p->elements;

    if (buf == NULL) {
      ENGINE->PerfError((void*) ENGINE, "vecdly: not initialised");
      return;
    }
    do {
      (*buf)[*indx] = *inVec++;
      fv1 = *indx - *dlyVec++ * GLOBLS->ekr;
      while (fv1 < 0.0f)      fv1 += (MYFLT)maxd;
      while (fv1 >= (MYFLT)maxd) fv1 -= (MYFLT)maxd;
      if (fv1 < maxd - 1) fv2 = fv1 + 1;
      else                fv2 = 0.0f;
      v1 = (long)fv1;
      v2 = (long)fv2;
      *outVec++ = (*buf)[v1] + (fv1 - v1) * ((*buf)[v2]-(*buf)[v1]);
      ++buf;
      if (++(*indx) == maxd) *indx = 0;
      ++indx;
    } while (--elements);
}

/* ----------------------------------------------- */

static void vseg_set(VSEG *p)
{
    TSEG    *segp;
    int     nsegs;
    MYFLT   **argp, dur, *vector;
    FUNC    *nxtfunc, *curfunc, *ftp;
    long    flength;

    {
      int   outocount, xoutcode, inocount, xincode;
      char  **strargs;
      ENGINE->GetOpcodeParams((void*) p, &outocount, &xoutcode,
                              &inocount, &xincode, &strargs);
      nsegs = ((inocount - 2) >> 1);        /* count segs & alloc if nec */
    }

    if ((segp = (TSEG *) p->auxch.auxp) == NULL) {
      ENGINE->AuxAlloc((void*) p, &p->auxch,
                       (size_t) ((nsegs + 1) * sizeof(TSEG)));
      p->cursegp = segp = (TSEG *) p->auxch.auxp;
      (segp+nsegs)->cnt = MAXPOS;
    }
    argp = p->argums;
    if ((nxtfunc = ENGINE->QueryFTableInit((void*) p, *argp++)) == NULL)
      return;
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ioutfunc)) != NULL) {
      p->vector = ftp->ftable;
      p->elements = (int) *p->ielements;
    }
    else
      return;
    if (p->elements >= ftp->flen) {
      ENGINE->InitError((void*) ENGINE,
                        "vlinseg/vexpseg: invalid num. of elements");
      return;
    }
    vector = p->vector;
    flength = p->elements;

    do {
      *vector++ = FL(0.0);
    } while (--flength);

    if (**argp <= FL(0.0))  return;     /* if idur1 <= 0, skip init  */
    p->cursegp = segp;                  /* else proceed from 1st seg */
    segp--;
    do {
      segp++;                   /* init each seg ..  */
      curfunc = nxtfunc;
      dur = **argp++;
      if ((nxtfunc = ENGINE->QueryFTableInit((void*) p, *argp++)) == NULL)
        return;
      if (dur > 0.0f) {
        segp->d = dur * GLOBLS->ekr;
        segp->function =  curfunc;
        segp->nxtfunction = nxtfunc;
        segp->cnt = (long) (segp->d + .5);
      }
      else
        break;                  /*  .. til 0 dur or done */
    } while (--nsegs);
    segp++;
    segp->d = 0.0f;
    segp->cnt = MAXPOS;         /* set last cntr to infin */
    segp->function =  nxtfunc;
    segp->nxtfunction = nxtfunc;
}

static void vlinseg(VSEG *p)
{
    TSEG    *segp;
    MYFLT   *curtab, *nxttab,curval, nxtval, durovercnt=0.0f, *vector;
    long    flength, upcnt;

    if (p->auxch.auxp == NULL) {
      ENGINE->PerfError((void*) ENGINE, "tableseg: not initialised");
      return;
    }
    segp = p->cursegp;
    curtab = segp->function->ftable;
    nxttab = segp->nxtfunction->ftable;
    upcnt = (long) segp->d-segp->cnt;
    if (upcnt > 0)
      durovercnt = segp->d/upcnt;
    while (--segp->cnt < 0)
      p->cursegp = ++segp;
    flength = p->elements;
    vector = p->vector;
    do {
      curval = *curtab++;
      nxtval = *nxttab++;
      if (durovercnt > 0.0)
        *vector++ = (curval + ((nxtval - curval) / durovercnt));
      else
        *vector++ = curval;
    } while (--flength);
}

static void vexpseg(VSEG *p)
{
    TSEG    *segp;
    MYFLT   *curtab, *nxttab,curval, nxtval, cntoverdur = FL(0.0), *vector;
    long    flength, upcnt;

    if (p->auxch.auxp == NULL) {
      ENGINE->PerfError((void*) ENGINE, "tablexseg: not initialised");
      return;
    }
    segp = p->cursegp;
    curtab = segp->function->ftable;
    nxttab = segp->nxtfunction->ftable;
    upcnt = (long) segp->d-segp->cnt;
    if (upcnt > 0)
      cntoverdur = upcnt/ segp->d;
    while (--segp->cnt < 0)
      p->cursegp = ++segp;
    flength = p->elements;
    vector = p->vector;
    cntoverdur *= cntoverdur;
    do {
      curval = *curtab++;
      nxtval = *nxttab++;
      *vector++ = (curval + ((nxtval - curval) * cntoverdur));
    } while (--flength);
}

/* ----------------------------------------------- */

static void vphaseseg_set(VPSEG *p)
{
    TSEG2   *segp;
    int     nsegs, j;
    MYFLT   **argp,  *vector;
    double  dur, durtot = 0.0, prevphs;
    FUNC    *nxtfunc, *curfunc, *ftp;
    long    flength;

    {
      int   outocount, xoutcode, inocount, xincode;
      char  **strargs;
      ENGINE->GetOpcodeParams((void*) p, &outocount, &xoutcode,
                              &inocount, &xincode, &strargs);
      nsegs = ((inocount - 3) >> 1);        /* count segs & alloc if nec */
    }

    if ((segp = (TSEG2 *) p->auxch.auxp) == NULL) {
      ENGINE->AuxAlloc((void*) p, &p->auxch,
                       (size_t) ((nsegs + 1) * sizeof(TSEG)));
      p->cursegp = segp = (TSEG2 *) p->auxch.auxp;
    }
    argp = p->argums;
    if ((nxtfunc = ENGINE->QueryFTableInit((void*) p, *argp++)) == NULL)
      return;
    if ((ftp = ENGINE->QueryFTableInit((void*) p, p->ioutfunc)) != NULL) {
      p->vector = ftp->ftable;
      p->elements = (int) *p->ielements;
    }
    else
      return;
    if ( p->elements > ftp->flen )
      ENGINE->InitError((void*) ENGINE, "vphaseseg: invalid num. of elements");
    vector = p->vector;
    flength = p->elements;

    do {
      *vector++ = FL(0.0);
    } while (--flength);

    if (**argp <= FL(0.0))              /* if idur1 <= 0, skip init  */
      return;                           /* else proceed from 1st seg */

    segp--;
    do {
      segp++;                 /* init each seg ..  */
      curfunc = nxtfunc;
      dur = **argp++;
      if ((nxtfunc = ENGINE->QueryFTableInit((void*) p, *argp++)) == NULL)
        return;
      if (dur > FL(0.0)) {
        durtot += dur;
        segp->d = dur;
        segp->function = curfunc;
        segp->nxtfunction = nxtfunc;
      }
      else break;             /*  .. til 0 dur or done */
    } while (--nsegs);
    segp++;

    segp->function =  nxtfunc;
    segp->nxtfunction = nxtfunc;
    nsegs = p->nsegs;

    segp = p->cursegp;

    for (j = 0; j < nsegs; j++)
      segp[j].d /= durtot;

    for (j = nsegs - 1; j >= 0; j--)
      segp[j + 1].d = segp[j].d;

    segp[0].d = prevphs = 0.0;

    for (j = 0; j <= nsegs; j++) {
      segp[j].d += prevphs;
      prevphs = segp[j].d;
    }
}

static void vphaseseg(VPSEG *p)
{
    TSEG2   *segp = p->cursegp;
    double  phase = (double) *p->kphase, partialPhase = 0.0;
    int     j, flength;
    MYFLT   *curtab = NULL, *nxttab = NULL, curval, nxtval, *vector;

    while (phase >= 1.0)
      phase -= 1.0;
    while (phase < 0.0)
      phase += 0.0;

    for (j = 0; j < p->nsegs; j++) {
      TSEG2 *seg = &segp[j], *seg1 = &segp[j + 1];
      if (phase < seg1->d) {
        curtab = seg->function->ftable;
        nxttab = seg1->function->ftable;
        partialPhase = (phase - seg->d) / (seg1->d - seg->d);
        break;
      }
    }
    if (!curtab)
      return;

    flength = p->elements;
    vector = p->vector;
    do {
      curval = *curtab++;
      nxtval = *nxttab++;
      *vector++ = (MYFLT) (curval + ((nxtval - curval) * partialPhase));
    } while (--flength);
}

void OPCODE(vectors_opcode_init) (void *csound)
{
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vadd",       sizeof(VECTOROP),   3,      "",     "iki",
      (SUBR) vectorOp_set,      (SUBR) vadd,    (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vmult",      sizeof(VECTOROP),   3,      "",     "iki",
      (SUBR) vectorOp_set,      (SUBR) vmult,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vpow",       sizeof(VECTOROP),   3,      "",     "iki",
      (SUBR) vectorOp_set,      (SUBR) vpow,    (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vexp",       sizeof(VECTOROP),   3,      "",     "iki",
      (SUBR) vectorOp_set,      (SUBR) vexp,    (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vaddv",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vaddv,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vsubv",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vsubv,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vmultv",     sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vmultv,  (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vdivv",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vdivv,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vpowv",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vpowv,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vexpv",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vexpv,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vcopy",      sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vcopy,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vmap",       sizeof(VECTORSOP),  3,      "",     "iii",
      (SUBR) vectorsOp_set,     (SUBR) vmap,    (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vlimit",     sizeof(VLIMIT),     3,      "",     "ikki",
      (SUBR) vlimit_set,        (SUBR) vlimit,  (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vwrap",      sizeof(VLIMIT),     3,      "",     "ikki",
      (SUBR) vlimit_set,        (SUBR) vwrap,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vmirror",    sizeof(VLIMIT),     3,      "",     "ikki",
      (SUBR) vlimit_set,        (SUBR) vmirror, (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vlinseg",    sizeof(VSEG),       3,      "",     "iin",
      (SUBR) vseg_set,          (SUBR) vlinseg, (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vexpseg",    sizeof(VSEG),       3,      "",     "iin",
      (SUBR) vseg_set,          (SUBR) vexpseg, (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vrandh",     sizeof(VRANDH),     3,      "",     "ikki",
      (SUBR) vrandh_set,        (SUBR) vrandh,  (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vrandi",     sizeof(VRANDI),     3,      "",     "ikki",
      (SUBR) vrandi_set,        (SUBR) vrandi,  (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vport",      sizeof(VPORT),      3,      "",     "ikio",
      (SUBR) vport_set,         (SUBR) vport,   (SUBR) NULL,    (SUBR) NULL);
    ((CSOUNDAPI*) csound)->AppendOpcode(csound,
      "vecdelay",   sizeof(VECDEL),     3,      "",     "iiiiio",
      (SUBR) vecdly_set,        (SUBR) vecdly,  (SUBR) NULL,    (SUBR) NULL);
}

