#!/bin/bash

# makedll.sh -- written by Istvan Varga, Feb 2003

# parse command line arguments

if ( [ $# -lt 4 ] ) ; then
	echo -e "usage: makedll.sh [-l | -p] -o outfile.dll OBJECTS [LIBS]"
	exit -1 ;
fi

if ( ( [ "$1" != "-l" ] ) && ( [ "$1" != "-p" ] ) ) ; then
	echo -e "usage: makedll.sh [-l | -p] -o outfile.dll OBJECTS [LIBS]"
	exit -1 ;
fi

if ( [ "$2" != "-o" ] ) ; then
	echo -e "usage: makedll.sh [-l | -p] -o outfile.dll OBJECTS [LIBS]"
	exit -1 ;
fi

STEM=`echo -e -n "$3" | cut -d "." -f 1`

MODE=$1
BASEFILE=$STEM.b
EXP_FILE=$STEM.e
LIB_FILE=lib$STEM.a
DLL_FILE=$STEM.dll
DEF_FILE=$STEM.def
shift 3

if ( ! ( [ -e "$DEF_FILE" ] ) ) ; then
	echo -e "makedll.sh: error: cannot find def file $DEF_FILE"
	exit -1 ;
fi

# programs to be used

AS=i686-pc-mingw32-as
LD=i686-pc-mingw32-gcc
DLLTOOL=i686-pc-mingw32-dlltool

rm -f $BASEFILE $EXP_FILE $DLL_FILE $LIB_FILE

$DLLTOOL -S $AS --dllname $DLL_FILE --def $DEF_FILE			\
	--output-lib $LIB_FILE

$LD -s -mdll -Wl,-O1 -Wl,--subsystem,windows -o $DLL_FILE		\
	-Wl,--base-file,$BASEFILE "$@"

$DLLTOOL -S $AS --dllname $DLL_FILE --base-file $BASEFILE		\
	--output-exp $EXP_FILE --def $DEF_FILE

$LD -s -mdll -Wl,-O1 -Wl,--subsystem,windows -o $DLL_FILE		\
	-Wl,--base-file,$BASEFILE "$@" $EXP_FILE

$DLLTOOL -S $AS --dllname $DLL_FILE --base-file $BASEFILE		\
	--output-exp $EXP_FILE --def $DEF_FILE

$LD -s -mdll -Wl,-O1 -Wl,--subsystem,windows -o $DLL_FILE		\
	"$@" $EXP_FILE

rm -f $BASEFILE
if ( [ "$MODE" != "-l" ] ) ; then
	rm -f $LIB_FILE $EXP_FILE ;
fi

